#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <errno.h>
#include "sm_alloc.h"
#include "gff3.h"

#define talloc(T) sm_talloc(T)

static struct gff3_blk *gff3_blk_new(size_t blk_size)
{
	struct gff3_blk *b;

	b = malloc(sizeof(struct gff3_blk) + (sizeof(gff3_rec_t) * blk_size));

	if (b == NULL)
		return (NULL);

	b->next     = NULL;
	b->blk_size = blk_size;
	b->blk_next = 0;

	memset(b->blk_item, 0, sizeof(gff3_rec_t) * blk_size);

	return (b);
}

static void gff3_rec_free_content(gff3_rec_t *r)
{
	register uint16_t i;

	free(r->seq_id);
	free(r->source);
	free(r->type);
	free(r->attr_tbl);

	return;
}

static void gff3_blk_free(struct gff3_blk *blk)
{
	size_t bsz, i;

	bsz = sizeof(struct gff3_blk) + (sizeof(gff3_rec_t) * blk->blk_size);

	for (i = 0; i < blk->blk_next; ++i)
		gff3_rec_free_content(blk->blk_item + i);

	memset(blk, 0, bsz);
	free(blk);

	return;
}

gff3_t *gff3_new(const char *path)
{
	gff3_t *g = talloc(gff3_t);

	if (g == NULL)
		return (NULL);

	if (strcmp(path, "-") == 0)
		g->fp = stdout;
	else
		g->fp = fopen(path, "w");

	if (g->fp == NULL) {
		free(g);
		return (NULL);
	}

	g->path = strdup(path);
	g->blk0 = gff3_blk_new(GFF3_DEFAULT_BLKSIZE);
	g->blkN = g->blk0;

	if (g->blk0 == NULL) {
		free(g);
		return (NULL);
	}

	g->rcnt      = 0;
	g->wrote_hdr = 0;

	return (g);
}

gff3_rec_t *gff3_rec_new(void)
{
	gff3_rec_t *r = talloc(gff3_rec_t);

	r->seq_id = NULL;
	r->source = NULL;
	r->type   = NULL;

	r->start  = 0;
	r->end    = 0;

	r->score  = 0.0;

	r->strand = '+';
	r->phase  =  -1;

	r->attr_cnt = 0;
	r->attr_tbl = NULL;

	return (r);
}

int gff3_rec_attr_add(gff3_rec_t *rec, char *tag, char *value)
{
	rec->attr_tbl = sm_realloc(rec->attr_tbl, sizeof(gff3_attr_t) * ++rec->attr_cnt);

	rec->attr_tbl[rec->attr_cnt - 1].tag   = tag;
	rec->attr_tbl[rec->attr_cnt - 1].value = value;

	return (0);
}

int gff3_add_raw(gff3_t *g, gff3_rec_t *rec)
{
	struct gff3_blk *dst;

	if (g == NULL) {
		errno = EFAULT;
		return -1;
	}

	if (g->blk0 == NULL) {
		g->blk0 = gff3_blk_new(GFF3_DEFAULT_BLKSIZE);
		g->blkN = g->blk0;
		dst     = g->blk0;

		goto blk_add;
	}

	dst = g->blk0;

	for (;;) {
		if (dst->blk_next < dst->blk_size)
			goto blk_add;

		if (dst->next == NULL) {
			dst->next = gff3_blk_new(GFF3_DEFAULT_BLKSIZE);
			g->blkN   = dst->next;
			dst       = g->blkN;

			goto blk_add;
		}

		dst = dst->next;
	}

	return (-1);
blk_add:
	if (dst == NULL) abort (); /* FIXME: use assume_d() */

	/*
	 * We don't duplicate the memory allocated for strings in the rec
	 * structure here.
	 */
	memcpy (dst->blk_item + dst->blk_next, rec, sizeof (gff3_rec_t));
	++dst->blk_next;
	++g->rcnt;

	return (0);
}

static int gff3_write_hdr(FILE *fp)
{
	rewind(fp);
	return !ferror(fp) ?
		fputs("##gff-version 3\n", fp) : -1;
}

static int gff3_write_rec(FILE *fp, gff3_rec_t *rec)
{
	char phase;

	switch (rec->phase) {
	case 0:  phase = '0'; break;
	case 1:  phase = '1'; break;
	case 2:	 phase = '2'; break;
	default: phase = '.';
	}

	fprintf(fp,
		"%s\t"         /* sequence ID */
		"%s\t"         /* source */
		"%s\t"         /* type (SOFA term or SOFA access #) */
		"%"PRIu64"\t"  /* start */
		"%"PRIu64"\t"  /* end */
		"%g\t"         /* score */
		"%c\t"         /* strand */
		"%c\t",        /* phase */
		rec->seq_id,
		rec->source,
		rec->type,
		rec->start,
		rec->end,
		rec->score,
		rec->strand,
		phase);

	if (rec->attr_cnt > 0) {
		register uint16_t i;

		for (i = 0; i < (rec->attr_cnt - 1); ++i)
			fprintf (fp, "%s=%s;", rec->attr_tbl[i].tag, rec->attr_tbl[i].value);
		if ((rec->attr_cnt - i) > 0)
			fprintf (fp, "%s=%s",  rec->attr_tbl[i].tag, rec->attr_tbl[i].value);
	} else
		fputc('.', fp);

	fprintf(fp, "\n");

	return (0);
}

int gff3_write(gff3_t *g)
{
	register struct gff3_blk *blk, *old;
	register size_t i;

	if (g == NULL) {
		errno = EFAULT;
		return (-1);
	}

	if (g->wrote_hdr == 0) {
		gff3_write_hdr(g->fp);
		g->wrote_hdr = 1;
	}

	blk = g->blk0;

	while (blk != NULL) {
		for (i = 0; i < blk->blk_next; ++i) {
			if (gff3_write_rec(g->fp, blk->blk_item + i) < 0)
				return (-1);
			/* TODO: add blk_offset for better error recovery */
		}

		old = blk;
		blk = blk->next;

		//gff3_blk_free(old);
		fflush(g->fp);
	}

	return(0);
}

int gff3_free(gff3_t *g)
{
	return (-1);
}
